/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
#include "server.cmdline.h"
#include "server.h"
#include "afs.h"
#include "send.h"
#include "list.h"
#include <ortp/ortp.h>
#include "ortp.h"

#define TARGET_ADDR(oc) inet_ntoa((oc)->addr)

extern struct gengetopt_args_info conf;

struct ortp_target {
	struct in_addr addr;
	int status, chunk_ts, last_ts;
	struct list_head node;
	int port, streaming;
	RtpSession *session;
};

static int numtargets;
static struct list_head targets;
static struct sender *self;

static void ortp_delete_target(struct ortp_target *ot, const char *msg)
{
	PARA_NOTICE_LOG("deleting %s:%d (%s) from list\n", TARGET_ADDR(ot),
		ot->port, msg);
	if (ot->session) {
		rtp_session_destroy(ot->session);
		ot->session = NULL;
	}
	numtargets--;
	list_del(&ot->node);
	free(ot);
	return;
}

static void ortp_send_buf(const char *buf, int len, long unsigned chunks_sent)
{
	struct ortp_target *ot, *tmp;
	int ret;

	list_for_each_entry_safe(ot, tmp, &targets, node) {
		struct timeval now;
		int ts;
		gettimeofday(&now, NULL);
		if (!ot->session)
			continue;
		ts = rtp_session_time_to_ts(ot->session, ot->chunk_ts)
			* chunks_sent;
		ret = rtp_session_send_with_ts(ot->session, buf, len, ts);
//		PARA_DEBUG_LOG("%lu:%lu: sent buf@%d (%d bytes), ret = %d, chunk_ts: %d\n",
//			now.tv_sec, now.tv_usec, ts, len, ret, ot->chunk_ts);
		ot->last_ts = ts;
		if (ret < 0)
			ortp_delete_target(ot, "send error");
		if (ret != len +12)
			PARA_NOTICE_LOG("short write %d\n", ret);
	}
}

static void ortp_send_packet(long unsigned chunks_sent, char type,
		const char *payload, size_t len)
{
	char *buf = para_malloc(3 + len);

	buf[0] = type;
	if (len)
		memcpy(buf + 3, payload, len);
	ortp_send_buf(buf, len + 3, chunks_sent);
	free(buf);
}

static void ortp_send_header(long unsigned chunks_sent, char *header,
	size_t header_len, const char *payload, size_t payload_len)
{
	char *buf;

	buf = para_malloc(header_len + 3 + payload_len);
	buf[0] = ORTP_HEADER;
	buf[1] = header_len >> 8;
	buf[2] = header_len % 256;
	PARA_DEBUG_LOG("sending header (3 + %d + %d bytes), buf=%x.%x\n", header_len,
		payload_len, buf[1] & 0xff, buf[2] & 0xff);
	memcpy(buf + 3, header, header_len);
	if (payload_len)
		memcpy(buf + 3 + header_len, payload, payload_len);
	ortp_send_buf(buf, header_len + payload_len + 3, chunks_sent);
	free(buf);
}

static void ortp_init_session(struct ortp_target *ot)
{
	RtpSession *s;
	int ret;

	PARA_NOTICE_LOG("sending to udp %s:%d\n", TARGET_ADDR(ot), ot->port);
	ot->session = rtp_session_new(RTP_SESSION_SENDONLY);
	if (!ot->session)
		return;
	s = ot->session;
//	rtp_session_set_jitter_compensation(ot->session, 100);
	/* always successful */
	rtp_session_set_send_payload_type(s, PAYLOAD_AUDIO_CONTINUOUS);
	ret = rtp_session_set_remote_addr(s, TARGET_ADDR(ot), ot->port);
	if (ret < 0) {
		rtp_session_destroy(ot->session);
		ot->session = NULL;
	}
}

/* called by afs */
static void ortp_shutdown_targets(void)
{
	char buf[2];
	struct ortp_target *ot, *tmp;

	buf[0] = ORTP_EOF;
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (!ot->session || !ot->streaming)
			continue;
		PARA_INFO_LOG("sending eof to ortp target %s:%d, ts = %d\n", TARGET_ADDR(ot), ot->port,
			ot->last_ts);
		rtp_session_send_with_ts(ot->session, buf, 1, ot->last_ts);
		ot->streaming = 0;
		ot->chunk_ts = 0;
		rtp_session_reset(ot->session);
	}
}

static int need_extra_header(struct audio_format *af, long unsigned chunks_sent)
{
	int mod = conf.ortp_header_interval_arg / (tv2ms(&af->chunk_tv) + 1);
	char *header;
	int len;

	if (!af || !af->get_header_info)
		return 0;
	header = af->get_header_info(&len);
	if (!header || len <= 0)
		return 0;
	if ((chunks_sent % mod))
		return 0;
	return 1;
}

static int ortp_send(struct audio_format *af, long unsigned current_chunk,
		long unsigned chunks_sent, const char *buf, size_t len)
{
	struct ortp_target *ot, *tmp;

	if (self->status != SENDER_ON)
		return 1;
//	PARA_DEBUG_LOG("len: %d, af: %p\n", len, af);
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (!ot->chunk_ts)
			ot->chunk_ts = tv2ms(&af->chunk_tv);
		if (!ot->session) {
			ortp_init_session(ot);
			if (!ot->session)
				continue;
		}
		ot->streaming = 1;
	}
	if (!current_chunk) {
		ortp_send_packet(chunks_sent, ORTP_BOF, buf, len);
		return 1;
	}
	if (need_extra_header(af, chunks_sent)) {
		int header_len;
		char *header = af->get_header_info(&header_len);
		ortp_send_header(chunks_sent, header, header_len, buf, len);
		return 1;
	}
	ortp_send_packet(chunks_sent, ORTP_DATA, buf, len);
	return 1;
}

static int ortp_com_on(struct sender_command_data *scd)
{
	self->status = SENDER_ON;
	return 1;
}

static int ortp_com_off(struct sender_command_data *scd)
{
	ortp_shutdown_targets();
	self->status = SENDER_OFF;
	return 1;
}

static int ortp_com_delete(struct sender_command_data *scd)
{
	char *a = para_strdup(inet_ntoa(scd->addr));
	struct ortp_target *ot, *tmp;
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (scd->port != ot->port)
			continue;
		if (strcmp(TARGET_ADDR(ot), a))
			continue;
		ortp_delete_target(ot, "com_delete");
	}
	return 1;
}

static void ortp_add_target(int port, struct in_addr *addr)
{
	struct ortp_target *ot = para_calloc(sizeof(struct ortp_target));
	ot->port = port;
	ot->addr = *addr;
	PARA_INFO_LOG("adding to target list (%s:%d)\n",
		TARGET_ADDR(ot), ot->port);
	list_add(&ot->node, &targets);
}

static int ortp_com_add(struct sender_command_data *scd)
{
	int port = (scd->port > 0)? scd->port : conf.ortp_default_port_arg;
	ortp_add_target(port, &scd->addr);
	return 1;
}

static char *ortp_info(void)
{
	struct ortp_target *ot;
	char *ret, *tgts = NULL;

	list_for_each_entry(ot, &targets, node) {
		char *tmp = make_message("%s%s:%d ", tgts? tgts : "",
			TARGET_ADDR(ot), ot->port);
		free(tgts);
		tgts = tmp;
	}
	ret = make_message(
		"ortp status: %s\n"
		"ortp default port: udp %d\n"
		"ortp targets: %s\n",
		(self->status == SENDER_ON)? "on" : "off",
		conf.ortp_default_port_arg,
		tgts? tgts : "(none)"
	);
	free(tgts);
	return ret;
}

static void ortp_init_target_list(void)
{
	int i;

	INIT_LIST_HEAD(&targets);
	for (i = 0; i < conf.ortp_target_given; i++) {
		char *arg = para_strdup(conf.ortp_target_arg[i]);
		char *p = strchr(arg, ':');
		int port;
		struct in_addr addr;

		if (!p)
			goto err;
		*p = '\0';
		if (!inet_aton(arg, &addr))
			goto err;
		port = atoi(++p);
		if (port < 0 || port > 65535)
			port = conf.ortp_default_port_arg;
		ortp_add_target(port, &addr);
		goto success;
err:
		PARA_CRIT_LOG("syntax error for ortp_target option "
			"#%d, ignoring\n", i);
success:
		free(arg);
		continue;
	}
}

static int ortp_set_fds(struct audio_format *af, int *max_fileno,
		fd_set *rfds, fd_set *wfds)
{
	return 1;
}

static char *ortp_help(void)
{
	return make_message(
		"usage: {on|off}\n"
		"usage: {add|delete} IP port\n"
		"example: add 224.0.1.38 1500 (LAN-streaming)\n"
	);
}

void para_ortp_init(struct sender *s)
{
	ortp_init();
	ortp_set_debug_file("oRTP", NULL);
	INIT_LIST_HEAD(&targets);
	s->info = ortp_info;
	s->help = ortp_help;
	s->send = ortp_send;
	s->set_fds = ortp_set_fds;
	s->handle_fds = NULL;
	s->shutdown_clients = ortp_shutdown_targets;
	s->client_cmds[SENDER_ON] = ortp_com_on;
	s->client_cmds[SENDER_OFF] = ortp_com_off;
	s->client_cmds[SENDER_DENY] = NULL;
	s->client_cmds[SENDER_ALLOW] = NULL;
	s->client_cmds[SENDER_ADD] = ortp_com_add;
	s->client_cmds[SENDER_DELETE] = ortp_com_delete;
	self = s;
	s->status = SENDER_OFF;
	ortp_init_target_list();
	if (!conf.ortp_no_autostart_given)
		s->status = SENDER_ON;
	PARA_DEBUG_LOG("%s", "ortp sender init complete\n");
}
